#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <mpi.h>

#define MINARG 6
#if !defined(TRUE)
#define TRUE 1
#define FALSE 0
#endif
#define MAXPROC 128
#define MAXTIMES 10000
#define MAXMSG 1024*1024
#define MAXTAG 8192

main(int argc, char *argv[]) {

int numero[MAXPROC];  
int MonNumero;
int n;
int inum, info, msg_rc;
int i,q,ii,iinorm,jj;
char *message;
char *message_rec;
double T, MaxT;
double time, time_final;
double time1, time2, barrtime, atime();
int taille,source,bufid;
int ntimes=MAXTIMES, number_of_tasks;
int nmax, P, K, barrflag, firstfirst=TRUE;
MPI_Status  status[2*MAXPROC];
MPI_Request request[2*MAXPROC];

FILE *output;


MPI_Init(&argc,&argv); /* initialize MPI environment */
MPI_Comm_rank(MPI_COMM_WORLD, &MonNumero); /* get task number */
MPI_Comm_size(MPI_COMM_WORLD,&number_of_tasks); /* get number of tasks */

if(argc<MINARG) {
   usage(argv[0]);
   MPI_Finalize(); exit(-1);
}

P=atoi(argv[1]);
if(P!=number_of_tasks) {
   if(!MonNumero) fprintf(stderr,"wrong number of tasks: must be %d\n",number_of_tasks);
   MPI_Finalize(); exit(-1);
}
K=atoi(argv[2]);
if(K>(P-1)) {
   if(!MonNumero) fprintf(stderr,"Too many comm: %d is the max\n",P-1);
   MPI_Finalize(); exit(-1);
}
nmax=atoi(argv[3]);
if(nmax>MAXMSG) {
   if(!MonNumero) fprintf(stderr,"Msg too long: %d is the max\n",MAXMSG);
   MPI_Finalize(); exit(-1);
}
ntimes=atoi(argv[4]);
if(ntimes>MAXTIMES) {
   if(!MonNumero) fprintf(stderr,"Too many times: %d is the max\n",MAXTIMES);
   MPI_Finalize(); exit(-1);
}

barrflag=atoi(argv[5]);

if(!(message=(char *)malloc(nmax))) {
   fprintf(stderr,"Could not allocate memory %d\n", nmax);
   MPI_Finalize(); exit(-1);
}

if(!(message_rec=(char*)malloc(nmax*(P+1)))) {
   fprintf(stderr,"Could not allocate memory %d\n", nmax*P);
   MPI_Finalize(); exit(-1);
}

if(!MonNumero) {
output = fopen("KTK_RES","w");
fprintf(output,"#============================================\n");
fprintf(output,"#KTK on %d PROCESSORS\n",P);
fprintf(output,"#============================================\n");
}

msg_rc = MPI_Bcast(&P,1,MPI_INT,0,MPI_COMM_WORLD);
  if(msg_rc < 0) {  
      fprintf(stderr,"error in second broadcast: %d\n",msg_rc);
      MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
  }
msg_rc = MPI_Bcast(&K,1,MPI_INT,0,MPI_COMM_WORLD);
  if(msg_rc < 0) {  
      fprintf(stderr,"error in second broadcast: %d\n",msg_rc);
      MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
  }
msg_rc = MPI_Bcast(&nmax,1,MPI_INT,0,MPI_COMM_WORLD);
  if(msg_rc < 0) {  
      fprintf(stderr,"error in second broadcast: %d\n",msg_rc);
      MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
  }
msg_rc = MPI_Bcast(&ntimes,1,MPI_INT,0,MPI_COMM_WORLD);
  if(msg_rc < 0) {  
      fprintf(stderr,"error in second broadcast: %d\n",msg_rc);
      MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
  }
msg_rc = MPI_Bcast(&barrflag,1,MPI_INT,0,MPI_COMM_WORLD);
  if(msg_rc < 0) {  
      fprintf(stderr,"error in second broadcast: %d\n",msg_rc);
      MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
  }

if(!MonNumero) for(ii=0; ii<P; ii++) numero[ii]=ii;

msg_rc = MPI_Bcast(numero,P,MPI_INT,0,MPI_COMM_WORLD);
  if(msg_rc < 0) {  
      fprintf(stderr,"error in second broadcast: %d\n",msg_rc);
      MPI_Abort(MPI_COMM_WORLD,-100); /* stop all other tasks */
  }

msg_rc=MPI_Barrier(MPI_COMM_WORLD);

if(!MonNumero) printf("Starting loop...\n");
time1=MPI_Wtime();
n = 1;
while(n <= nmax)
 {
  barrtime=time=0;
  for(i = 0; i < ntimes; i++)
     {

      if(barrflag)
         msg_rc=MPI_Barrier(MPI_COMM_WORLD);

      time1=MPI_Wtime();
      for(ii = P-1+MonNumero, jj=0; ii > (P-1-K)+MonNumero; ii--, jj++) {
	  iinorm=ii%number_of_tasks;
          msg_rc=MPI_Irecv(&message_rec[n*iinorm], n, MPI_BYTE, iinorm,
                            n%MAXTAG, MPI_COMM_WORLD, request+jj);
      }

      for(ii = MonNumero+1; ii < K+MonNumero+1; ii++,jj++) {
	  iinorm=ii%number_of_tasks;
          msg_rc=MPI_Isend(&message[0], n, MPI_BYTE, iinorm,
                              n%MAXTAG, MPI_COMM_WORLD, request+jj);
      }
     
      msg_rc=MPI_Waitall(2*K,request,status);

      time2 = MPI_Wtime();
      MaxT = time2-time1;

      if(barrflag) {
         msg_rc=MPI_Barrier(MPI_COMM_WORLD);
         barrtime+=(MPI_Wtime()-time2);
      }

      MPI_Reduce(&MaxT,&T,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

      if(!MonNumero) time = time + T;
     }

  if(!MonNumero) {
     time_final = time/ntimes;
 
     printf("%6d  %8.6f %f\n",n,(time_final/2.), (double)(n*K*2)*1.e-6/time_final);
     if(barrflag) {
#if 0
        printf("mean time in barrier %f\n",barrtime/ntimes); 
        fprintf(output,"time in barrier %f\n",barrtime/ntimes); 
#endif
     }
     fprintf(output,
             "%7d %8.6f %f\n", n,(time_final/2.), (double)(n*K*2)/time_final);
  }

  n = n*2;
}

if(!MonNumero) fclose(output);


msg_rc=MPI_Barrier(MPI_COMM_WORLD);
MPI_Finalize();
exit(0);


}

usage(char *prog) {
  
      fprintf(stderr,
"Usage: %s <# of tasks (<%d)> <# of comm (<%d)> <max len (< %d)> <max times (<%d)> <barrier flag>\n", prog,MAXPROC,MAXPROC-1,MAXMSG,MAXTIMES);
}
