/************************************************************************
 *	Read COPYRIGHT file in this directory				*
 ************************************************************************/

/*
 *  $Id: chinit.c,v 1.2 1998/03/13 22:32:27 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/* 
    This file contains the routines that provide the basic information 
    on the device, and initialize it
 */

/* We put these include FIRST incase we are building the memory debugging
   version; since these includes may define malloc etc., we need to include 
   them before mpid.h 
 */
#ifndef HAVE_STDLIB_H
extern char *getenv();
#else
#include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "mpid.h"
#include "mpiddev.h"
#include "mpimem.h"
#include <stdio.h>
#include "chgamma.h"
#include "threadsafe.h"

/* #define DEBUG(a) {a} */
#define DEBUG(a)


/*  Home of globals  */

AG_pkt_q_entry		*AG_pkt_queue;
unsigned long		AG_head, AG_tail;
AG_queue_entry		*AG_src_queue;
AG_buf			*AG_buffer;
void			*AG_tempbuf_ptr;
MPID_PKT_HEAD_T		AG_pkt;
time_586		AG_time_origin, AG_time;

#ifdef AG_HARDMEMLOCK
AG_lock_table		AG_lock;
#endif

#ifdef AG_FLOWCTL
AG_flow_table_entry	*AG_flow;
#endif

unsigned long		AGR_send_id;
AG_buf			*AGR_buffer;

time_586 a,b;


/* Prototypes  */

int MPID_CH_End ANSI_ARGS(( MPID_Device * ));
int MPID_CH_Abort ANSI_ARGS(( struct MPIR_COMMUNICATOR *, int, char * ));
void MPID_CH_Version_name ANSI_ARGS(( char * ));


/* Init/Exit/Abort functions  */

/* 
Returns a device.  
This sets up a message-passing device (short/eager/rendezvous protocols)
*/

MPID_Device *MPID_CH_InitMsgPass(
	int *argc, char ***argv, int short_len, int long_len
) {
	MPID_Device *dev;
	int i;

	dev = (MPID_Device *)MALLOC( sizeof(MPID_Device) );
	if (!dev) return 0;
	/* The short protocol is for messages no longer than AG_SHORTSIZE */
	if ((short_len < 0) || (short_len > AG_SHORTSIZE))
		 short_len = AG_SHORTSIZE;
	if ((long_len < 0) || (long_len > AG_EAGRSIZE))
		long_len  = AG_EAGRSIZE;
	dev->long_len     = short_len+1;
	dev->vlong_len    = long_len+1;
	/*  We use the same protocol for short as well as eager  */
	dev->short_msg    = MPID_CH_Short_setup();
	dev->long_msg     = MPID_CH_Short_setup();
	dev->vlong_msg    = MPID_CH_Rndvb_setup();
	dev->eager        = dev->long_msg;
	dev->rndv         = dev->vlong_msg;
	dev->check_device = MPID_CH_Check_incoming;
	dev->terminate    = MPID_CH_End;
	dev->abort	      = MPID_CH_Abort;
	dev->next	      = 0;

	/*
	Set the file for Debugging output.  The actual output is controlled
	by MPIDDebugFlag
	*/
#ifdef MPID_DEBUG_ALL
	if (MPID_DEBUG_FILE == 0) MPID_DEBUG_FILE = stdout;
#endif

	gamma_init_argv(argc,argv);

	MPID_MyWorldRank = gamma_my_node();
	MPID_MyWorldSize = gamma_how_many_nodes();

	AG_head = 0;
	AG_tail = 0;
	gamma_mlock(&AG_head, sizeof(AG_head));
	gamma_mlock(&AG_tail, sizeof(AG_tail));
	AG_spinlock_init();

	AG_src_queue = (AG_queue_entry *)
		malloc(AG_MAXQLEN * sizeof(AG_queue_entry));
	gamma_mlock(AG_src_queue, AG_MAXQLEN * sizeof(AG_queue_entry));

	AG_pkt_queue = (AG_pkt_q_entry *)
		malloc(AG_MAXQLEN * sizeof(AG_pkt_q_entry));
	gamma_mlock(AG_pkt_queue,
		AG_MAXQLEN * sizeof(AG_pkt_q_entry));

	gamma_mlock(&AG_pkt, sizeof(MPID_PKT_HEAD_T));

	AG_buffer = (AG_buf *)malloc(MPID_MyWorldSize * sizeof(AG_buf));
	gamma_mlock(AG_buffer,
		MPID_MyWorldSize * sizeof(AG_buf));
	for (i=0;i<MPID_MyWorldSize;i++)
		if (i != gamma_my_node()) {
			AG_buffer[i] = (AG_buf)malloc(sizeof(MPID_PKT_HEAD_T));
			gamma_mlock(AG_buffer[i],sizeof(MPID_PKT_HEAD_T));
			/*
			 *  For data messages and ctl messages
			 *  except REQUEST_SEND:
			 *  Connect port "i" to port "me" of i-th
			 *  remote processor.
			 *  Port "i" will receive data from i-th processor
			 *  into the temp buffer pointed by AG_buffer[i].
			 */
			gamma_set_active_port_2p(i,
				i, gamma_my_par_pid(), gamma_my_node(),
				AG_handler_rec,
				i,
				GO_BACK, AG_buffer[i],
				sizeof(MPID_PKT_HEAD_T),
				sizeof(MPID_PKT_HEAD_T)
			);
		}
#ifdef AG_HARDMEMLOCK
	AG_lock.how_many = 0;
	AG_lock.head = 0;
	AG_lock.table =
		(AG_lock_entry *)malloc(AG_MAXLOCK * sizeof(AG_lock_entry));
	for (i=0;i<AG_MAXLOCK;i++)
		AG_lock.table[i].in_use = 0;
#endif

#ifdef AG_FLOWCTL
	AG_flow = (AG_flow_table_entry *)
		malloc(MPID_MyWorldSize * sizeof(AG_flow_table_entry));
	for (i=0;i<MPID_MyWorldSize;i++) {
		if (i == MPID_MyWorldRank) continue;
		AG_flow[i].msgs = AG_MAXQLEN/(MPID_MyWorldSize-1);
		AG_flow[i].padding = 0;
	}
	gamma_mlock(AG_flow, MPID_MyWorldSize * sizeof(AG_flow_table_entry));
#endif

	AGR_buffer = (AG_buf *)malloc(MPID_MyWorldSize * sizeof(AG_buf));
	gamma_mlock(AGR_buffer,MPID_MyWorldSize * sizeof(AG_buf));
	for (i=0;i<MPID_MyWorldSize;i++)
		if (i != gamma_my_node()) {
			AGR_buffer[i] = (AG_buf)
				malloc(sizeof(MPID_PKT_T));
			gamma_mlock(AGR_buffer[i],
				sizeof(MPID_PKT_T));
			/*
			 *  For REQUEST_SEND ctl messages:
			 *  Connect port "i+<no. of nodes>" to port
			 *  "me+<no. of nodes>" of i-th remote processor.
			 *  Port "i+<no. of nodes>" will receive REQUEST_SEND
			 *  ctl messages from i-th processor
			 *  into the temp buffer pointed by AGR_buffer[i].
			 */
			gamma_set_active_port(i+MPID_MyWorldSize,
				i, gamma_my_par_pid(),
				gamma_my_node()+MPID_MyWorldSize,
				AGR_handler_rec,
				i,
				GO_BACK, AGR_buffer[i],
				sizeof(MPID_PKT_T)
			);
		}

	gamma_sync();

	/*
	 *  Set the time origin.  This should happen ``simultaneously''
	 *  at all processors, however there is no guarantee that it will
	 *  be so, although a barrier sync. has just been issued.
	 */
	gamma_time(AG_time_origin);

	DEBUG_PRINT_MSG("Finished init");

	MPID_DO_HETERO(MPID_CH_Init_hetero( argc, argv ));

    DEBUG_PRINT_MSG("Leaving MPID_CH_InitMsgPass");

    return dev;
}


int MPID_CH_Abort( comm_ptr, code, msg )
struct MPIR_COMMUNICATOR *comm_ptr;
int      code;
char     *msg;
{
    if (msg) {
	fprintf( stderr, "[%d] %s\n", MPID_MyWorldRank, msg );
    }
    else {
	fprintf( stderr, "[%d] Aborting program!\n", MPID_MyWorldRank );
    }
    fflush( stderr );
    fflush( stdout );
    gamma_munlockall();
    gamma_abort();
    exit(1);
}


int MPID_CH_End(MPID_Device *dev)
{
    DEBUG_PRINT_MSG("Entering MPID_CH_End\n");
    /* Finish off any pending transactions */
    /* MPID_CH_Complete_pending(); */

    if (MPID_GetMsgDebugFlag()) {
	MPID_PrintMsgDebug();
    }
#ifdef MPID_HAS_HETERO /* #HETERO_START# */
    MPID_CH_Hetero_free();
#endif                 /* #HETERO_END# */
    (dev->short_msg->delete)( dev->short_msg );
    (dev->long_msg->delete)( dev->long_msg );
    (dev->vlong_msg->delete)( dev->vlong_msg );
    FREE( dev );

    /* We should really generate an error or warning message if there 
       are uncompleted operations... */
    
    gamma_munlockall();
/* peppe
gamma_time(AG_time);
if (!MPID_MyWorldRank)
	fprintf(stderr,"MPI/GAMMA: Time from MPI_Init() to MPI_Exit() %g sec\n",
		gamma_time_diff(AG_time,AG_time_origin)*1e-6);
*/
    gamma_exit();
    return 0;
}


void MPID_CH_Version_name( name )
char *name;
{
	sprintf( name, "ADI/GAMMA version %4.2f\n",MPIDPATCHLEVEL);
}

